use std::collections::HashMap;
use std::convert::Infallible;
use std::net::SocketAddr;
use std::result::Result;

use hyper::{Body, Method, Request, Response, Server, StatusCode};
use hyper::service::{make_service_fn, service_fn};
pub use mysql_async::*;
pub use mysql_async::prelude::*;
use serde::{Deserialize, Serialize};

fn get_url() -> String {
    if let Ok(url) = std::env::var("DATABASE_URL") {
        let opts = Opts::from_url(&url).expect("DATABASE_URL invalid");
        if opts
            .db_name()
            .expect("a database name is required")
            .is_empty()
        {
            panic!("database name is empty");
        }
        url
    } else {
        "mysql://root:pass@127.0.0.1:3306/mysql".into()
    }
}

#[derive(Serialize, Deserialize, Debug)]
struct Todo {
    id: i32,
    title: String,
    completed: bool,
}

#[derive(Serialize, Deserialize, Debug)]
struct CreateTodo {
    title: String,
    completed: bool,
}

impl Todo {
    fn new(id: i32, title: String, completed: bool) -> Self {
        Self {
            id,
            title,
            completed,
        }
    }
}

async fn handle_request(req: Request<Body>, pool: Pool) -> Result<Response<Body>, anyhow::Error> {
    match (req.method(), req.uri().path()) {
        (&Method::GET, "/") => Ok(Response::new(Body::from(
            "The valid endpoints are /init /create_todo /update_todo /todos /delete_todo",
        ))),

        // Simply echo the body back to the client.
        (&Method::POST, "/echo") => Ok(Response::new(req.into_body())),

        // CORS OPTIONS
        (&Method::OPTIONS, "/init") => Ok(response_build(&String::from(""))),
        (&Method::OPTIONS, "/create_todo") => Ok(response_build(&String::from(""))),
        (&Method::OPTIONS, "/update_todo") => Ok(response_build(&String::from(""))),
        (&Method::OPTIONS, "/delete_todo") => Ok(response_build(&String::from(""))),
        (&Method::OPTIONS, "/todos") => Ok(response_build(&String::from(""))),

        (&Method::GET, "/init") => {
            let mut conn = pool.get_conn().await.unwrap();
            "DROP TABLE IF EXISTS todos;".ignore(&mut conn).await?;
            "CREATE TABLE todos (id INT AUTO_INCREMENT, title VARCHAR(255), completed BOOLEAN, constraint todos_pk primary key (id));".ignore(&mut conn).await?;
            drop(conn);
            Ok(response_build("{\"status\":true}"))
        }

        (&Method::POST, "/create_todo") => {
            let mut conn = pool.get_conn().await.unwrap();

            let byte_stream = hyper::body::to_bytes(req).await?;
            let todo: CreateTodo = serde_json::from_slice(&byte_stream).unwrap();

            "INSERT INTO todos (title, completed) VALUES (:title, :completed)"
                .with(params! {
                    "title" => &todo.title,
                    "completed" => todo.completed,
                })
                .ignore(&mut conn)
                .await?;


            drop(conn);
            Ok(response_build("{\"status\":true}"))
            // Ok(Response::new(Body::from("{\"status\":true}")))
        }

        (&Method::POST, "/update_todo") => {
            let mut conn = pool.get_conn().await.unwrap();

            let byte_stream = hyper::body::to_bytes(req).await?;
            let todo: Todo = serde_json::from_slice(&byte_stream).unwrap();

            "UPDATE todos SET title = :title, completed = :completed WHERE id = :id"
                .with(params! {
                    "id" => todo.id,
                    "title" => &todo.title,
                    "completed" => todo.completed,
                })
                .ignore(&mut conn)
                .await?;
            drop(conn);
            Ok(response_build("{\"status\":true}"))
            // Ok(Response::new(Body::from("{\"status\":true}")))
        }

        (&Method::GET, "/todos") => {
            let mut conn = pool.get_conn().await.unwrap();
            let todos = "SELECT * FROM todos;"
                .with(())
                .map(&mut conn, |(id, title, completed)| {
                    Todo::new(
                        id,
                        title,
                        completed,
                    )
                })
                .await?;
            drop(conn);
            Ok(response_build(&serde_json::to_string(&todos)?))
        }

        (&Method::GET, "/delete_todo") => {
            let mut conn = pool.get_conn().await.unwrap();

            let params: HashMap<String, String> = req.uri().query().map(|v| {
                url::form_urlencoded::parse(v.as_bytes()).into_owned().collect()
            }).unwrap_or_else(HashMap::new);
            let todo_id = params.get("id");

            "DELETE FROM todos WHERE id=:id"
                .with(params! { "id" => todo_id, })
                .ignore(&mut conn)
                .await?;

            drop(conn);
            Ok(response_build("{\"status\":true}"))
            // Ok(Response::new(Body::from("{\"status\":true}")))
        }

        // Return the 404 Not Found for other routes.
        _ => {
            let mut not_found = Response::default();
            *not_found.status_mut() = StatusCode::NOT_FOUND;
            Ok(not_found)
        }
    }
}

// CORS headers
fn response_build(body: &str) -> Response<Body> {
    Response::builder()
        .header("Access-Control-Allow-Origin", "*")
        .header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")
        .header("Access-Control-Allow-Headers", "*")
        .body(Body::from(body.to_owned()))
        .unwrap()
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> Result<(), Box<dyn std::error::Error + Send + Sync>> {
    let opts = Opts::from_url(&*get_url()).unwrap();
    let builder = OptsBuilder::from_opts(opts);
    // The connection pool will have a min of 5 and max of 10 connections.
    let constraints = PoolConstraints::new(5, 10).unwrap();
    let pool_opts = PoolOpts::default().with_constraints(constraints);
    let pool = Pool::new(builder.pool_opts(pool_opts));

    let addr = SocketAddr::from(([0, 0, 0, 0], 8080));
    let make_svc = make_service_fn(|_| {
        let pool = pool.clone();
        async move {
            Ok::<_, Infallible>(service_fn(move |req| {
                let pool = pool.clone();
                handle_request(req, pool)
            }))
        }
    });
    let server = Server::bind(&addr).serve(make_svc);
    if let Err(e) = server.await {
        eprintln!("server error: {}", e);
    }
    Ok(())

    /*
    let listener = TcpListener::bind(addr).await?;
    println!("Listening on http://{}", addr);
    loop {
        let (stream, _) = listener.accept().await?;

        tokio::task::spawn(async move {
            if let Err(err) = Http::new().serve_connection(stream, service_fn(handle_request)).await {
                println!("Error serving connection: {:?}", err);
            }
        });
    }
    */
}
